/*
 * Created on Oct 31, 2003
 */
package isac.util.parser;

import isac.util.CalcChanged;
import isac.util.Message;
import isac.util.ResponseWrapper;
import isac.util.formulae.Assumptions;
import isac.util.formulae.CalcHead;
import isac.util.formulae.CalcHeadCompoundID;
import isac.util.formulae.CalcHeadItem;
import isac.util.formulae.CalcHeadItemList;
import isac.util.formulae.CalcHeadSimpleID;
import isac.util.formulae.FormHeadsContainer;
import isac.util.formulae.Formula;
import isac.util.formulae.CalcFormula;
import isac.util.formulae.Position;
import isac.util.formulae.Specification;
import isac.util.tactics.Rewrite;
import isac.util.tactics.RewriteInst;
import isac.util.tactics.RewriteSet;
import isac.util.tactics.RewriteSetInst;
import isac.util.tactics.SimpleTactic;
import isac.util.tactics.StringListTactic;
import isac.util.tactics.SubProblemTactic;
import isac.util.tactics.TacticsContainer;
import isac.util.tactics.Theorem;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;

import org.apache.commons.digester.Digester;
import org.apache.log4j.Logger;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * parses Objects from XML-Output of the Kernel.
 * 
 * @author Richard Gradischnegg
 */
public class XMLParserDigest {

	private Digester digester;

	private String dtdPath;

	class BridgeErrorHandler implements ErrorHandler {
		public void error(SAXParseException e) {
			System.out.println("---xml-error " + e);
		}

		public void fatalError(SAXParseException e) {
			System.out.println("---xml-fatalerror " + e);
		}

		public void warning(SAXParseException e) {
			System.out.println("---xml-warning " + e);
		}
	}

	// auxillary class for feeding the dtd
	class BridgeEntityResolver implements EntityResolver {
		String dtd;

		public BridgeEntityResolver(String dtd) {
			this.dtd = dtd;
		}

		public InputSource resolveEntity(String publicId, String systemId)
				throws SAXException, IOException {
			if ((systemId != null) && systemId.endsWith("isac.dtd")) {
				//System.out.println("------->inside entity resolv");
				return new InputSource(new StringReader(dtd));
				//new StringReader("<!ELEMENT dummy EMPTY>"));
			} else {
				System.out.println("BridgeEntityResolver: return null");
				return null;
			}
		}
	}

	static Logger logger = Logger.getLogger(XMLParserDigest.class.getName());

	/**
	 * Constructor: create a new XMLParserDigest object and fill it with the
	 * rules.
	 * 
	 * @param dtdPath
	 *            Path to the DTD-File (Document Type Description)
	 */
	public XMLParserDigest(String dtdPath) {
		logger.debug("XMLParserDigest: dtdPath=" + dtdPath);
		digester = new Digester();
		digester.setErrorHandler(new BridgeErrorHandler());
		this.dtdPath = dtdPath;
		//dtd-File --> String (faster validating)
		StringBuffer sb;
		try {
			BufferedReader br = new BufferedReader(new FileReader(dtdPath));
			sb = new StringBuffer();
			while (br.ready()) {
				sb.append(br.readLine() + "\n");
			}
			// Set dtd for parser validation
			digester.setEntityResolver(new BridgeEntityResolver(sb.toString()));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		// Rules for the digester to create Objects and fill them from XML-Files
		digester.addObjectCreate("ISAC", ResponseWrapper.class);
		digester.addObjectCreate("*/CALCID", CalcHeadSimpleID.class);
		digester.addCallMethod("*/CALCID", "setID", 0);
		digester.addSetRoot("*/CALCID", "setCalcID");

		// Rules for parsing Position
		digester.addObjectCreate("*/POSITION", Position.class);
		digester.addCallMethod("*/POSITION/INTLIST/INT", "addInt", 0);
		digester.addCallMethod("*/POSITION/POS", "setKind", 0);
		digester.addSetNext("*/POSITION", "setResponse");
//		digester.addSetNext("*/POSITION", "setPosition");

		// 3 different Positions in CalcChangedEvent
		digester.addObjectCreate("*/UNCHANGED", Position.class);
		digester.addCallMethod("*/UNCHANGED/INTLIST/INT", "addInt", 0);
		digester.addCallMethod("*/UNCHANGED/POS", "setKind", 0);
		digester.addSetNext("*/UNCHANGED", "setLastUnchanged");

		digester.addObjectCreate("*/DELETED", Position.class);
		digester.addCallMethod("*/DELETED/INTLIST/INT", "addInt", 0);
		digester.addCallMethod("*/DELETED/POS", "setKind", 0);
		digester.addSetNext("*/DELETED", "setLastDeleted");

		digester.addObjectCreate("*/GENERATED", Position.class);
		digester.addCallMethod("*/GENERATED/INTLIST/INT", "addInt", 0);
		digester.addCallMethod("*/GENERATED/POS", "setKind", 0);
		digester.addSetNext("*/GENERATED", "setLastGenerated");

		// Rules for parsing Formula
		digester.addObjectCreate("*/FORMULA", Formula.class);
		digester.addCallMethod("*/FORMULA/MATHML/ISA", "setText", 0);
		digester.addSetNext("*/FORMULA", "setFormula");

		// Rules for parsing CalcFormula with Position
		digester.addObjectCreate("*/CALCFORMULA", CalcFormula.class);
		digester.addSetNext("*/CALCFORMULA", "setResponse");

		//	Rules for parsing CalcHead
		digester.addObjectCreate("*/CALCHEAD", CalcHead.class);
		digester.addSetProperties("*/CALCHEAD", "status", "statusString");

		digester.addCallMethod("*/CALCHEAD/HEAD/MATHML/ISA", "setHead", 0);
		digester.addCallMethod("*/CALCHEAD/BELONGSTO", "setBelongsTo", 0);
		addItemList(digester, "Given");
		addItemList(digester, "Where");
		addItemList(digester, "Find");
		addItemList(digester, "Relate");
		digester.addObjectCreate("*/CALCHEAD/SPECIFICATION",
				Specification.class);
		digester.addObjectCreate("*/CALCHEAD/SPECIFICATION/THEORYID",
				CalcHeadSimpleID.class);
		digester.addCallMethod("*/CALCHEAD/SPECIFICATION/THEORYID", "setID", 0);
		digester.addSetNext("*/CALCHEAD/SPECIFICATION/THEORYID", "setTheory");
		digester.addObjectCreate("*/CALCHEAD/SPECIFICATION/PROBLEMID",
				CalcHeadCompoundID.class);
		digester.addCallMethod(
				"*/CALCHEAD/SPECIFICATION/PROBLEMID/STRINGLIST/STRING",
				"addString", 0);
		digester.addSetNext("*/CALCHEAD/SPECIFICATION/PROBLEMID", "setProblem");
		digester.addObjectCreate("*/CALCHEAD/SPECIFICATION/METHODID",
				CalcHeadCompoundID.class);
		digester.addCallMethod(
				"*/CALCHEAD/SPECIFICATION/METHODID/STRINGLIST/STRING",
				"addString", 0);
		digester.addSetNext("*/CALCHEAD/SPECIFICATION/METHODID", "setMethod");
		digester.addSetNext("*/CALCHEAD/SPECIFICATION", "setSpecification");
		digester.addSetNext("*/CALCHEAD", "setResponse");

		//	Rules for parsing various Tactic Objects
		//	Rules for parsing SimpleTactic
		digester.addObjectCreate("*/SIMPLETACTIC", SimpleTactic.class);
		digester.addCallMethod("*/SIMPLETACTIC/MATHML/ISA", "setArgument", 0);
		digester.addSetProperties("*/SIMPLETACTIC", "name", "name");
		digester.addSetNext("*/SIMPLETACTIC", "setResponse");

		//	Rules for parsing KeyListTactic
		digester.addObjectCreate("*/STRINGLISTTACTIC", StringListTactic.class);
		digester.addCallMethod("*/STRINGLISTTACTIC/STRINGLIST/STRING",
				"addKey", 0);
		digester.addSetProperties("*/STRINGLISTTACTIC", "name", "name");
		digester.addSetNext("*/STRINGLISTTACTIC", "setResponse");

		//  Rules for parsing Theorem
		digester.addObjectCreate("*/THEOREM", Theorem.class);
		digester.addCallMethod("*/THEOREM/ID", "setId", 0);
		digester.addCallMethod("*/THEOREM/FORMULA/MATHML/ISA", "setFormula", 0);
		digester.addSetNext("*/THEOREM", "setTheorem");

		//  Rules for parsing RewriteTactic
		digester.addObjectCreate("*/REWRITETACTIC", Rewrite.class);
		digester.addSetProperties("*/REWRITETACTIC", "name", "name");
		digester.addSetNext("*/REWRITETACTIC", "setResponse");

		//  Rules for parsing RewriteInstTactic
		digester.addObjectCreate("*/REWRITEINSTTACTIC", RewriteInst.class);
		digester.addSetProperties("*/REWRITEINSTTACTIC", "name", "name");
		digester.addSetNext("*/REWRITEINSTTACTIC", "setResponse");
		digester.addCallMethod(
				"*/REWRITEINSTTACTIC/SUBSTITUTION/PAIR/VARIABLE/MATHML/ISA",
				"setVariable",//WN0501 ^^^ this seems to work only on spec.case
				0);
		digester.addCallMethod(
				"*/REWRITEINSTTACTIC/SUBSTITUTION/PAIR/VALUE/MATHML/ISA",
				"setValue",//WN0501 ^^^ this seems to work only on spec.case
				0);

		//	Rules for parsing RewriteSetTactic
		digester.addObjectCreate("*/REWRITESETTACTIC", RewriteSet.class);
		digester.addCallMethod("*/REWRITESETTACTIC/RULESET", "setRuleSet", 0);
		digester.addSetProperties("*/REWRITESETTACTIC", "name", "name");
		digester.addSetNext("*/REWRITESETTACTIC", "setResponse");

		//	Rules for parsing RewriteSetInstTactic
		digester
				.addObjectCreate("*/REWRITESETINSTTACTIC", RewriteSetInst.class);
		digester.addCallMethod("*/REWRITESETINSTTACTIC/RULESET", "setRuleSet",
				0);
		digester.addCallMethod(
				"*/REWRITESETINSTTACTIC/SUBSTITUTION/PAIR/VARIABLE/MATHML/ISA",
				"setVariable",//WN0501 ^^^ this seems to work only on spec.case
				0);
		digester.addCallMethod(
				"*/REWRITESETINSTTACTIC/SUBSTITUTION/PAIR/VALUE/MATHML/ISA",
				"setValue",//WN0501 ^^^ this seems to work only on spec.case
				0);
		digester.addSetProperties("*/REWRITESETINSTTACTIC", "name", "name");
		digester.addSetNext("*/REWRITESETINSTTACTIC", "setResponse");

		// Rules for parsing SubProblemTactic
		digester.addObjectCreate("*/SUBPROBLEMTACTIC", SubProblemTactic.class);
		digester.addCallMethod("*/SUBPROBLEMTACTIC/THEORY", "setTheoryID", 0);
		digester.addCallMethod("*/SUBPROBLEMTACTIC/PROBLEM/STRINGLIST/STRING",
				"addProblemString", 0);
		digester.addSetProperties("*/SUBPROBLEMTACTIC", "name", "name");
		digester.addSetNext("*/SUBPROBLEMTACTIC", "setResponse");

		//	Rules for parsing CalcIterator move message
		digester.addObjectCreate("*/CALCITERATOR/ERROR", Message.class);
		digester.addCallMethod("*/CALCITERATOR/ERROR", "setError", 0);
		digester.addSetNext("*/CALCITERATOR/ERROR", "setResponse");

		//	Rules for parsing CalcIterator constructor message
		digester.addObjectCreate("*/ADDUSER", CalcHeadSimpleID.class);
		digester.addCallMethod("*/ADDUSER/USERID", "setID", 0);
		digester.addSetNext("*/ADDUSER", "setResponse");

		//  Rules for parsing autoCalculateResponse
		digester.addCallMethod("*/AUTOCALC/ERROR", "setError", 0);
		digester.addObjectCreate("*/CALCCHANGED", CalcChanged.class);
		digester.addSetNext("*/CALCCHANGED", "setResponse");

		// Rules for parsing fetchApplicableTactics
		digester.addObjectCreate("*/APPLICABLETACTICS/TACLIST",
				TacticsContainer.class);
		digester.addSetNext("*/APPLICABLETACTICS/TACLIST", "setResponse");

		// Rules for getElementsFromTo
		digester.addObjectCreate("*/GETELEMENTSFROMTO/FORMHEADS",
				FormHeadsContainer.class);
		digester.addSetNext("*/GETELEMENTSFROMTO/FORMHEADS", "setResponse");

		// getAssumptions (and still getAccumulatedAsms)
		digester.addObjectCreate("*/ASMLIST",
				Assumptions.class);
		digester.addSetNext("*/ASMLIST", "setResponse");
	}

	private void addItemList(Digester digester, String modPart) {
		String modPartUp = modPart.toUpperCase();
		// Rules for parsing ItemLists in the model part of a CalcHead: Given
		// Find Where Relate
		digester.addObjectCreate("*/CALCHEAD/MODEL/" + modPartUp,
				CalcHeadItemList.class);
		digester.addObjectCreate("*/CALCHEAD/MODEL/" + modPartUp + "/ITEM",
				CalcHeadItem.class);
		digester.addSetProperties("*/CALCHEAD/MODEL/" + modPartUp + "/ITEM",
				"status", "itemStatus");
		digester.addCallMethod("*/CALCHEAD/MODEL/" + modPartUp
				+ "/ITEM/MATHML/ISA", "setText", 0);
		digester.addSetNext("*/CALCHEAD/MODEL/" + modPartUp + "/ITEM",
				"addItem");
		digester.addSetNext("*/CALCHEAD/MODEL/" + modPartUp, "set" + modPart);
	}

	/**
	 * Set whether or not using the DTD for validating
	 * 
	 * @param validating
	 *            <UL>
	 *            <LI>true: XML is tested for validity against DTD
	 *            <LI>
	 *            <LI>false: XML is only tested for well-formed-ness
	 *            <LI>
	 *            </UL>
	 */
	public void setValidating(boolean validating) {
		logger.warn("setValidating: validating=" + validating);
		digester.setValidating(validating);
	}

	/**
	 * Parse an XML string to Java Object(s)
	 * 
	 * @param xmlString
	 *            the string containing the XML content
	 * @return ResponseWrapper containing the CalcID and the parsed Object
	 */
	public ResponseWrapper parse(String xmlString) {
		Object o = null;
		xmlString = "<?xml version=\"1.0\"?><!DOCTYPE ISAC SYSTEM \"" + dtdPath
				+ "\"><ISAC>" + xmlString + "</ISAC>";

		encodeEntities(xmlString);
		try {
			digester.clear();
			//System.out.println("root === " + digester.getRoot());
			o = digester.parse(new InputSource(new StringReader(xmlString)));
			//System.out.println("--------------->" + o);
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		}
		ResponseWrapper wrapper = (ResponseWrapper) o;
		//System.out.println(wrapper.getCalcID().toSMLString());
		return wrapper;
	}

	public static void encodeEntities(String xmlString) {
		xmlString = xmlString.replaceAll(" & ", "&amp;");
		xmlString = xmlString.replaceAll(" > ", "&gt;");
		xmlString = xmlString.replaceAll(" < ", "&lt;");
	}

	public static void decodeEntities(String xmlString) {
		xmlString = xmlString.replaceAll("&amp;", " & ");
		xmlString = xmlString.replaceAll("&gt;", " > ");
		xmlString = xmlString.replaceAll("&lt;", " < ");
	}
}